# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.deprecation import RenamedClass
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.indexed_component import rule_wrapper
from pyomo.core.base.expression import (
    Expression,
    ExpressionData,
    ScalarExpression,
    IndexedExpression,
)
from pyomo.dae.contset import ContinuousSet
from pyomo.dae.diffvar import DAE_Error


@ModelComponentFactory.register("Integral Expression in a DAE model.")
class Integral(Expression):
    """
    Represents an integral over a continuous domain

    The :py:class:`Integral<pyomo.dae.Integral>` component can be used to
    represent an integral taken over the entire domain of a
    :py:class:`ContinuousSet<pyomo.dae.ContinuousSet>`. Once every
    :py:class:`ContinuousSet<pyomo.dae.ContinuousSet>` in a model has been
    discretized, any integrals in the model will be converted to algebraic
    equations using the trapezoid rule. Future development will include more
    sophisticated numerical integration methods.

    Parameters
    ----------
    *args
        Every indexing set needed to evaluate the integral expression

    wrt : :py:class:`ContinuousSet<pyomo.dae.ContinuousSet>`
        The continuous domain over which the integral is being taken

    rule : function
        Function returning the expression being integrated
    """

    def __new__(cls, *args, **kwds):
        if cls != Integral:
            return super(Integral, cls).__new__(cls)
        if len(args) == 0:
            raise ValueError("Integral must be indexed by a ContinuousSet")
        elif len(args) == 1:
            return ScalarIntegral.__new__(ScalarIntegral)
        else:
            return IndexedIntegral.__new__(IndexedIntegral)

    def __init__(self, *args, **kwds):
        if "wrt" in kwds and "withrespectto" in kwds:
            raise TypeError("Cannot specify both 'wrt' and 'withrespectto keywords")

        wrt = kwds.pop('wrt', None)
        wrt = kwds.pop('withrespectto', wrt)

        if wrt is None:
            # Check to be sure Integral is indexed by single
            # ContinuousSet and take Integral with respect to that
            # ContinuousSet
            if len(args) != 1:
                raise ValueError(
                    "Integral indexed by multiple ContinuousSets. "
                    "The desired ContinuousSet must be specified using the "
                    "keyword argument 'wrt'"
                )
            wrt = args[0]

        if type(wrt) is not ContinuousSet:
            raise ValueError(
                "Cannot take the integral with respect to '%s'. Must take an "
                "integral with respect to a ContinuousSet" % wrt
            )
        self._wrt = wrt

        loc = None
        for i, s in enumerate(args):
            if s is wrt:
                loc = i

        # Check that the wrt ContinuousSet is in the argument list
        if loc is None:
            raise ValueError(
                "The ContinuousSet '%s' was not found in the indexing sets "
                "of the Integral" % wrt.name
            )
        self.loc = loc

        # Remove the index that the integral is being expanded over
        arg = args[0:loc] + args[loc + 1 :]

        # Check that if bounds are given
        bounds = kwds.pop('bounds', None)
        if bounds is not None:
            raise DAE_Error(
                "Setting bounds on integrals has not yet been implemented. "
                "Integrals may only be taken over an entire ContinuousSet"
            )

        # Create integral expression and pass to the expression initialization
        intexp = kwds.pop('expr', None)
        intexp = kwds.pop('rule', intexp)
        if intexp is None:
            raise ValueError("Must specify an integral expression")

        _is_indexed = bool(len(arg))

        def _trap_rule(rule, m, *a):
            ds = sorted(m.find_component(wrt.local_name))
            return sum(
                0.5
                * (ds[i + 1] - ds[i])
                * (
                    rule(m, *(a[0:loc] + (ds[i + 1],) + a[loc:]))
                    + rule(m, *(a[0:loc] + (ds[i],) + a[loc:]))
                )
                for i in range(len(ds) - 1)
            )

        # Note that position_map is mapping arguments (block, *args), so
        # must be 1 more than len(args), and loc has to be offset by one
        kwds['rule'] = rule_wrapper(
            intexp,
            _trap_rule,
            positional_arg_map=(i for i in range(len(args) + 1) if i != loc + 1),
        )
        kwds.setdefault('ctype', Integral)
        Expression.__init__(self, *arg, **kwds)

    def get_continuousset(self):
        """Return the :py:class:`ContinuousSet<pyomo.dae.ContinuousSet>`
        the integral is being taken over
        """
        return self._wrt


class ScalarIntegral(ScalarExpression, Integral):
    """
    An integral that will have no indexing sets after applying a numerical
    integration transformation
    """

    def __init__(self, *args, **kwds):
        ExpressionData.__init__(self, None, component=self)
        Integral.__init__(self, *args, **kwds)

    def clear(self):
        self._data = {}

    def is_fully_discretized(self):
        """
        Checks to see if all ContinuousSets indexing this Integral have been
        discretized
        """
        if 'scheme' not in self._wrt.get_discretization_info():
            return False
        return True


class SimpleIntegral(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarIntegral
    __renamed__version__ = '6.0'


class IndexedIntegral(IndexedExpression, Integral):
    """
    An integral that will be indexed after applying a numerical integration
    transformation
    """

    def is_fully_discretized(self):
        """
        Checks to see if all ContinuousSets indexing this Integral have been
        discretized.
        """
        wrt = self._wrt
        if 'scheme' not in wrt.get_discretization_info():
            return False

        setlist = []
        if self.dim() == 1:
            setlist = [self.index_set()]
        else:
            setlist = self.index_set().set_tuple

        for i in setlist:
            if i.ctype is ContinuousSet:
                if 'scheme' not in i.get_discretization_info():
                    return False
        return True



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import enum
from .diff_with_sympy import differentiate as sympy_diff
from .diff_with_pyomo import reverse_sd, reverse_ad


class Modes(str, enum.Enum):
    sympy = 'sympy'
    reverse_symbolic = 'reverse_symbolic'
    reverse_numeric = 'reverse_numeric'

    # Overloading __str__ is needed to match the behavior of the old
    # pyutilib.enum class (removed June 2020). There are spots in the
    # code base that expect the string representation for items in the
    # enum to not include the class name. New uses of enum shouldn't
    # need to do this.
    def __str__(self):
        return self.value


def differentiate(expr, wrt=None, wrt_list=None, mode=Modes.reverse_numeric):
    """Return derivative of expression.

    This function returns the derivative of expr with respect to one or
    more variables.  The type of the return value depends on the
    arguments wrt, wrt_list, and mode. See below for details.

    Parameters
    ----------
    expr: pyomo.core.expr.numeric_expr.NumericExpression
        The expression to differentiate
    wrt: pyomo.core.base.var.VarData
        If specified, this function will return the derivative with
        respect to wrt. wrt is normally a VarData, but could
        also be a ParamData. wrt and wrt_list cannot both be specified.
    wrt_list: list of pyomo.core.base.var.VarData
        If specified, this function will return the derivative with
        respect to each element in wrt_list.  A list will be returned
        where the values are the derivatives with respect to the
        corresponding entry in wrt_list.
    mode: pyomo.core.expr.calculus.derivatives.Modes
        Specifies the method to use for differentiation. Should be one
        of the members of the Modes enum:

            Modes.sympy:
                The pyomo expression will be converted to a sympy
                expression. Differentiation will then be done with
                sympy, and the result will be converted back to a pyomo
                expression.  The sympy mode only does symbolic
                differentiation. The sympy mode requires exactly one of
                wrt and wrt_list to be specified.
            Modes.reverse_symbolic:
                Symbolic differentiation will be performed directly with
                the pyomo expression in reverse mode. If neither wrt nor
                wrt_list are specified, then a ComponentMap is returned
                where there will be a key for each node in the
                expression tree, and the values will be the symbolic
                derivatives.
            Modes.reverse_numeric:
                Numeric differentiation will be performed directly with
                the pyomo expression in reverse mode. If neither wrt nor
                wrt_list are specified, then a ComponentMap is returned
                where there will be a key for each node in the
                expression tree, and the values will be the floating
                point values of the derivatives at the current values of
                the variables.

    Returns
    -------
    res: float, :py:class:`NumericExpression`, :py:class:`ComponentMap`, or list
        The value or expression of the derivative(s)

    """

    try:
        mode = Modes(mode)
    except:
        raise ValueError(
            f'differentiate(): Unrecognized differentiation mode: {mode}\n'
            f'Expected one of {list(map(str, Modes))}.'
        )
    if mode == Modes.reverse_numeric or mode == Modes.reverse_symbolic:
        if mode == Modes.reverse_numeric:
            res = reverse_ad(expr=expr)
        else:
            res = reverse_sd(expr=expr)

        if wrt is not None:
            if wrt_list is not None:
                raise ValueError(
                    'differentiate(): Cannot specify both wrt and wrt_list.'
                )
            if wrt in res:
                res = res[wrt]
            else:
                res = 0
        elif wrt_list is not None:
            _res = list()
            for _wrt in wrt_list:
                if _wrt in res:
                    _res.append(res[_wrt])
                else:
                    _res.append(0)
            res = _res
    else:
        assert mode == Modes.sympy
        res = sympy_diff(expr=expr, wrt=wrt, wrt_list=wrt_list)

    return res


differentiate.Modes = Modes



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core import Var
from pyomo.core.base import Constraint
from pyomo.core.expr.compare import assertExpressionsEqual

# This file contains check methods shared between GDP inner representation-based
# transformations. Currently, those are the inner_representation_gdp and
# nested_inner_repn_gdp transformations, since each have disjuncts with the
# same structure.


# Check one disjunct from the log model for proper contents
def check_log_disjunct(test, d, pts, f, substitute_var, x):
    test.assertEqual(len(d.component_map(Constraint)), 3)
    # lambdas and indicator_var
    test.assertEqual(len(d.component_map(Var)), 2)
    test.assertIsInstance(d.lambdas, Var)
    test.assertEqual(len(d.lambdas), 2)
    for lamb in d.lambdas.values():
        test.assertEqual(lamb.lb, 0)
        test.assertEqual(lamb.ub, 1)
    test.assertIsInstance(d.convex_combo, Constraint)
    assertExpressionsEqual(test, d.convex_combo.expr, d.lambdas[0] + d.lambdas[1] == 1)
    test.assertIsInstance(d.set_substitute, Constraint)
    assertExpressionsEqual(
        test, d.set_substitute.expr, substitute_var == f(x), places=7
    )
    test.assertIsInstance(d.linear_combo, Constraint)
    test.assertEqual(len(d.linear_combo), 1)
    assertExpressionsEqual(
        test, d.linear_combo[0].expr, x == pts[0] * d.lambdas[0] + pts[1] * d.lambdas[1]
    )


# Check one disjunct from the paraboloid model for proper contents.
def check_paraboloid_disjunct(test, d, pts, f, substitute_var, x1, x2):
    test.assertEqual(len(d.component_map(Constraint)), 3)
    # lambdas and indicator_var
    test.assertEqual(len(d.component_map(Var)), 2)
    test.assertIsInstance(d.lambdas, Var)
    test.assertEqual(len(d.lambdas), 3)
    for lamb in d.lambdas.values():
        test.assertEqual(lamb.lb, 0)
        test.assertEqual(lamb.ub, 1)
    test.assertIsInstance(d.convex_combo, Constraint)
    assertExpressionsEqual(
        test, d.convex_combo.expr, d.lambdas[0] + d.lambdas[1] + d.lambdas[2] == 1
    )
    test.assertIsInstance(d.set_substitute, Constraint)
    assertExpressionsEqual(
        test, d.set_substitute.expr, substitute_var == f(x1, x2), places=7
    )
    test.assertIsInstance(d.linear_combo, Constraint)
    test.assertEqual(len(d.linear_combo), 2)
    assertExpressionsEqual(
        test,
        d.linear_combo[0].expr,
        x1
        == pts[0][0] * d.lambdas[0]
        + pts[1][0] * d.lambdas[1]
        + pts[2][0] * d.lambdas[2],
    )
    assertExpressionsEqual(
        test,
        d.linear_combo[1].expr,
        x2
        == pts[0][1] * d.lambdas[0]
        + pts[1][1] * d.lambdas[1]
        + pts[2][1] * d.lambdas[2],
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import pyomo.common.unittest as unittest

from pyomo.environ import Var, Set, ConcreteModel, TransformationFactory, pyomo
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.dae.diffvar import DAE_Error

from pyomo.repn import generate_standard_repn

from io import StringIO

from pyomo.common.log import LoggingIntercept

from os.path import abspath, dirname, normpath, join

currdir = dirname(abspath(__file__))
exdir = normpath(join(currdir, '..', '..', '..', 'examples', 'dae'))


def repn_to_rounded_dict(repn, digits):
    temp = dict()
    for i, v in enumerate(repn.linear_vars):
        temp[id(v)] = round(repn.linear_coefs[i], digits)
    return temp


class TestCollocation(unittest.TestCase):
    """
    Class for testing the pyomo.DAE collocation discretization
    """

    def setUp(self):
        """
        Setting up testing model
        """
        self.m = m = ConcreteModel()
        m.t = ContinuousSet(bounds=(0, 10))
        m.v1 = Var(m.t)
        m.dv1 = DerivativeVar(m.v1)
        m.s = Set(initialize=[1, 2, 3], ordered=True)

    # test collocation discretization with radau points
    # on var indexed by single ContinuousSet
    def test_disc_single_index_radau(self):
        m = self.m.clone()
        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=5, ncp=3)

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertTrue(len(m.dv1_disc_eq) == 15)
        self.assertTrue(len(m.v1) == 16)

        expected_tau_points = [0.0, 0.1550510257216822, 0.64494897427831788, 1.0]
        expected_disc_points = [
            0,
            0.310102,
            1.289898,
            2.0,
            2.310102,
            3.289898,
            4.0,
            4.310102,
            5.289898,
            6.0,
            6.310102,
            7.289898,
            8.0,
            8.310102,
            9.289898,
            10,
        ]
        disc_info = m.t.get_discretization_info()

        self.assertTrue(disc_info['scheme'] == 'LAGRANGE-RADAU')

        for idx, val in enumerate(disc_info['tau_points']):
            self.assertAlmostEqual(val, expected_tau_points[idx])

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])
            self.assertTrue(type(val) in [float, int])

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertTrue(m._pyomo_dae_reclassified_derivativevars[0] is m.dv1)

        repn_baseline = {
            id(m.dv1[2.0]): 1.0,
            id(m.v1[0]): 1.5,
            id(m.v1[0.310102]): -2.76599,
            id(m.v1[1.289898]): 3.76599,
            id(m.v1[2.0]): -2.5,
        }

        repn = generate_standard_repn(m.dv1_disc_eq[2.0].body)
        repn_gen = repn_to_rounded_dict(repn, 5)
        self.assertEqual(repn_baseline, repn_gen)

        repn_baseline = {
            id(m.dv1[4.0]): 1.0,
            id(m.v1[2.0]): 1.5,
            id(m.v1[2.310102]): -2.76599,
            id(m.v1[3.289898]): 3.76599,
            id(m.v1[4.0]): -2.5,
        }

        repn = generate_standard_repn(m.dv1_disc_eq[4.0].body)
        repn_gen = repn_to_rounded_dict(repn, 5)
        self.assertEqual(repn_baseline, repn_gen)

    # test collocation discretization with radau points
    # second order derivative indexed by single ContinuousSet
    def test_disc_second_order_radau(self):
        m = self.m.clone()
        m.dv1dt2 = DerivativeVar(m.v1, wrt=(m.t, m.t))
        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=2, ncp=2)

        self.assertTrue(hasattr(m, 'dv1dt2_disc_eq'))
        self.assertTrue(len(m.dv1dt2_disc_eq) == 4)
        self.assertTrue(len(m.v1) == 5)

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertTrue(m.dv1 in m._pyomo_dae_reclassified_derivativevars)
        self.assertTrue(m.dv1dt2 in m._pyomo_dae_reclassified_derivativevars)

        repn_baseline = {
            id(m.dv1dt2[5.0]): 1,
            id(m.v1[0]): -0.24,
            id(m.v1[1.666667]): 0.36,
            id(m.v1[5.0]): -0.12,
        }

        repn = generate_standard_repn(m.dv1dt2_disc_eq[5.0].body)
        repn_gen = repn_to_rounded_dict(repn, 5)
        self.assertEqual(repn_baseline, repn_gen)

        repn_baseline = {
            id(m.dv1dt2[10]): 1,
            id(m.v1[5.0]): -0.24,
            id(m.v1[6.666667]): 0.36,
            id(m.v1[10]): -0.12,
        }

        repn = generate_standard_repn(m.dv1dt2_disc_eq[10.0].body)
        repn_gen = repn_to_rounded_dict(repn, 5)
        self.assertEqual(repn_baseline, repn_gen)

    # test second order derivative with single collocation point
    def test_disc_second_order_1cp(self):
        m = ConcreteModel()
        m.t = ContinuousSet(bounds=(0, 1))
        m.t2 = ContinuousSet(bounds=(0, 10))
        m.v = Var(m.t, m.t2)
        m.dv = DerivativeVar(m.v, wrt=(m.t, m.t2))
        TransformationFactory('dae.collocation').apply_to(m, nfe=2, ncp=1)

        self.assertTrue(hasattr(m, 'dv_disc_eq'))
        self.assertTrue(len(m.dv_disc_eq) == 4)
        self.assertTrue(len(m.v) == 9)

    # test collocation discretization with legendre points
    # on var indexed by single ContinuousSet
    def test_disc_single_index_legendre(self):
        m = self.m.clone()
        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=5, ncp=3, scheme='LAGRANGE-LEGENDRE')

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertTrue(hasattr(m, 'v1_t_cont_eq'))
        self.assertTrue(len(m.dv1_disc_eq) == 15)
        self.assertTrue(len(m.v1_t_cont_eq) == 5)
        self.assertTrue(len(m.v1) == 21)

        expected_tau_points = [
            0.0,
            0.11270166537925834,
            0.49999999999999989,
            0.88729833462074226,
        ]
        expected_disc_points = [
            0,
            0.225403,
            1.0,
            1.774597,
            2.0,
            2.225403,
            3.0,
            3.774597,
            4.0,
            4.225403,
            5.0,
            5.774597,
            6.0,
            6.225403,
            7.0,
            7.774597,
            8.0,
            8.225403,
            9.0,
            9.774597,
            10,
        ]
        disc_info = m.t.get_discretization_info()

        self.assertTrue(disc_info['scheme'] == 'LAGRANGE-LEGENDRE')

        for idx, val in enumerate(disc_info['tau_points']):
            self.assertAlmostEqual(val, expected_tau_points[idx])

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])
            self.assertTrue(type(val) in [float, int])

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertTrue(m.dv1 in m._pyomo_dae_reclassified_derivativevars)

        repn_baseline = {
            id(m.dv1[3.0]): 1,
            id(m.v1[2.0]): -1.5,
            id(m.v1[2.225403]): 2.86374,
            id(m.v1[3.0]): -1.0,
            id(m.v1[3.774597]): -0.36374,
        }

        repn = generate_standard_repn(m.dv1_disc_eq[3.0].body)
        repn_gen = repn_to_rounded_dict(repn, 5)
        self.assertEqual(repn_baseline, repn_gen)

        repn_baseline = {
            id(m.dv1[5.0]): 1,
            id(m.v1[4.0]): -1.5,
            id(m.v1[4.225403]): 2.86374,
            id(m.v1[5.0]): -1.0,
            id(m.v1[5.774597]): -0.36374,
        }

        repn = generate_standard_repn(m.dv1_disc_eq[5.0].body)
        repn_gen = repn_to_rounded_dict(repn, 5)
        self.assertEqual(repn_baseline, repn_gen)

    # test collocation discretization with legendre points
    # second order derivative indexed by single ContinuousSet
    def test_disc_second_order_legendre(self):
        m = self.m.clone()
        m.dv1dt2 = DerivativeVar(m.v1, wrt=(m.t, m.t))
        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=2, ncp=2, scheme='LAGRANGE-LEGENDRE')

        self.assertTrue(hasattr(m, 'dv1dt2_disc_eq'))
        self.assertTrue(hasattr(m, 'v1_t_cont_eq'))
        self.assertTrue(len(m.dv1dt2_disc_eq) == 4)
        self.assertTrue(len(m.v1_t_cont_eq) == 2)
        self.assertTrue(len(m.v1) == 7)

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertTrue(m.dv1 in m._pyomo_dae_reclassified_derivativevars)
        self.assertTrue(m.dv1dt2 in m._pyomo_dae_reclassified_derivativevars)

        repn_baseline = {
            id(m.dv1dt2[1.056624]): 1,
            id(m.v1[0]): -0.48,
            id(m.v1[1.056624]): 0.65569,
            id(m.v1[3.943376]): -0.17569,
        }

        repn = generate_standard_repn(m.dv1dt2_disc_eq[1.056624].body)
        repn_gen = repn_to_rounded_dict(repn, 5)
        self.assertEqual(repn_baseline, repn_gen)

        repn_baseline = {
            id(m.dv1dt2[6.056624]): 1,
            id(m.v1[5.0]): -0.48,
            id(m.v1[6.056624]): 0.65569,
            id(m.v1[8.943376]): -0.17569,
        }

        repn = generate_standard_repn(m.dv1dt2_disc_eq[6.056624].body)
        repn_gen = repn_to_rounded_dict(repn, 5)
        self.assertEqual(repn_baseline, repn_gen)

    # test collocation discretization on var indexed by ContinuousSet and Set
    def test_disc_multi_index(self):
        m = self.m.clone()
        m.v2 = Var(m.t, m.s)
        m.dv2 = DerivativeVar(m.v2)

        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=5, ncp=3)

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertTrue(hasattr(m, 'dv2_disc_eq'))
        self.assertTrue(len(m.dv2_disc_eq) == 45)
        self.assertTrue(len(m.v2) == 48)

        expected_tau_points = [0.0, 0.1550510257216822, 0.64494897427831788, 1.0]
        expected_disc_points = [
            0,
            0.310102,
            1.289898,
            2.0,
            2.310102,
            3.289898,
            4.0,
            4.310102,
            5.289898,
            6.0,
            6.310102,
            7.289898,
            8.0,
            8.310102,
            9.289898,
            10,
        ]
        disc_info = m.t.get_discretization_info()

        self.assertTrue(disc_info['scheme'] == 'LAGRANGE-RADAU')

        for idx, val in enumerate(disc_info['tau_points']):
            self.assertAlmostEqual(val, expected_tau_points[idx])

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertTrue(m.dv1 in m._pyomo_dae_reclassified_derivativevars)
        self.assertTrue(m.dv2 in m._pyomo_dae_reclassified_derivativevars)

    # test collocation discretization on var indexed by multiple ContinuousSets
    def test_disc_multi_index2(self):
        m = self.m.clone()
        m.t2 = ContinuousSet(bounds=(0, 5))
        m.v2 = Var(m.t, m.t2)
        m.dv2dt = DerivativeVar(m.v2, wrt=m.t)
        m.dv2dt2 = DerivativeVar(m.v2, wrt=m.t2)

        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=2, ncp=2)

        self.assertTrue(hasattr(m, 'dv2dt_disc_eq'))
        self.assertTrue(hasattr(m, 'dv2dt2_disc_eq'))
        self.assertTrue(len(m.dv2dt_disc_eq) == 20)
        self.assertTrue(len(m.dv2dt2_disc_eq) == 20)
        self.assertTrue(len(m.v2) == 25)

        expected_t_disc_points = [0, 1.666667, 5.0, 6.666667, 10]
        expected_t2_disc_points = [0, 0.833333, 2.5, 3.333333, 5]

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_t_disc_points[idx])

        for idx, val in enumerate(list(m.t2)):
            self.assertAlmostEqual(val, expected_t2_disc_points[idx])

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertTrue(m.dv1 in m._pyomo_dae_reclassified_derivativevars)
        self.assertTrue(m.dv2dt in m._pyomo_dae_reclassified_derivativevars)
        self.assertTrue(m.dv2dt2 in m._pyomo_dae_reclassified_derivativevars)

    # test collocation discretization on var indexed by ContinuousSet
    # and multi-dimensional Set
    def test_disc_multidimen_index(self):
        m = self.m.clone()
        m.s2 = Set(initialize=[('A', 'B'), ('C', 'D'), ('E', 'F')])
        m.v2 = Var(m.t, m.s2)
        m.dv2 = DerivativeVar(m.v2)
        m.v3 = Var(m.s2, m.t)
        m.dv3 = DerivativeVar(m.v3)

        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=5, ncp=3)

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertTrue(hasattr(m, 'dv2_disc_eq'))
        self.assertTrue(hasattr(m, 'dv3_disc_eq'))
        self.assertTrue(len(m.dv2_disc_eq) == 45)
        self.assertTrue(len(m.v2) == 48)
        self.assertTrue(len(m.dv3_disc_eq) == 45)
        self.assertTrue(len(m.v3) == 48)

        expected_tau_points = [0.0, 0.1550510257216822, 0.64494897427831788, 1.0]
        expected_disc_points = [
            0,
            0.310102,
            1.289898,
            2.0,
            2.310102,
            3.289898,
            4.0,
            4.310102,
            5.289898,
            6.0,
            6.310102,
            7.289898,
            8.0,
            8.310102,
            9.289898,
            10,
        ]
        disc_info = m.t.get_discretization_info()

        self.assertTrue(disc_info['scheme'] == 'LAGRANGE-RADAU')

        for idx, val in enumerate(disc_info['tau_points']):
            self.assertAlmostEqual(val, expected_tau_points[idx])

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertTrue(m.dv1 in m._pyomo_dae_reclassified_derivativevars)
        self.assertTrue(m.dv2 in m._pyomo_dae_reclassified_derivativevars)
        self.assertTrue(m.dv3 in m._pyomo_dae_reclassified_derivativevars)

    # test passing the discretization invalid options
    def test_disc_invalid_options(self):
        m = self.m.clone()

        with self.assertRaises(TypeError):
            TransformationFactory('dae.collocation').apply_to(m, wrt=m.s)

        with self.assertRaises(ValueError):
            TransformationFactory('dae.collocation').apply_to(m, nfe=-1)

        with self.assertRaises(ValueError):
            TransformationFactory('dae.collocation').apply_to(m, ncp=0)

        with self.assertRaises(ValueError):
            TransformationFactory('dae.collocation').apply_to(m, scheme='foo')

        with self.assertRaises(ValueError):
            TransformationFactory('dae.collocation').apply_to(m, foo=True)

        TransformationFactory('dae.collocation').apply_to(m, wrt=m.t)
        with self.assertRaises(ValueError):
            TransformationFactory('dae.collocation').apply_to(m, wrt=m.t)

        m = self.m.clone()
        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m)
        with self.assertRaises(ValueError):
            disc.apply_to(m)

    # test looking up radau collocation points
    def test_lookup_radau_collocation_points(self):
        # Save initial flag value
        colloc_numpy_avail = pyomo.dae.plugins.colloc.numpy_available

        # Numpy flag must be False to test lookup
        pyomo.dae.plugins.colloc.numpy_available = False

        m = self.m.clone()
        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=5, ncp=3)

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertTrue(len(m.dv1_disc_eq) == 15)
        self.assertTrue(len(m.v1) == 16)

        expected_tau_points = [0.0, 0.1550510257216822, 0.64494897427831788, 1.0]
        expected_disc_points = [
            0,
            0.310102,
            1.289898,
            2.0,
            2.310102,
            3.289898,
            4.0,
            4.310102,
            5.289898,
            6.0,
            6.310102,
            7.289898,
            8.0,
            8.310102,
            9.289898,
            10,
        ]
        disc_info = m.t.get_discretization_info()

        self.assertTrue(disc_info['scheme'] == 'LAGRANGE-RADAU')

        for idx, val in enumerate(disc_info['tau_points']):
            self.assertAlmostEqual(val, expected_tau_points[idx])

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])

        m = self.m.clone()
        with self.assertRaises(ValueError):
            disc = TransformationFactory('dae.collocation')
            disc.apply_to(m, ncp=15, scheme='LAGRANGE-RADAU')

        # Restore initial flag value
        pyomo.dae.plugins.colloc.numpy_available = colloc_numpy_avail

    # test looking up legendre collocation points
    def test_lookup_legendre_collocation_points(self):
        # Save initial flag value
        colloc_numpy_avail = pyomo.dae.plugins.colloc.numpy_available

        # Numpy flag must be False to test lookup
        pyomo.dae.plugins.colloc.numpy_available = False

        m = self.m.clone()
        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=5, ncp=3, scheme='LAGRANGE-LEGENDRE')

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertTrue(len(m.dv1_disc_eq) == 15)
        self.assertTrue(len(m.v1) == 21)

        expected_tau_points = [
            0.0,
            0.11270166537925834,
            0.49999999999999989,
            0.88729833462074226,
        ]
        expected_disc_points = [
            0,
            0.225403,
            1.0,
            1.774597,
            2.0,
            2.225403,
            3.0,
            3.774597,
            4.0,
            4.225403,
            5.0,
            5.774597,
            6.0,
            6.225403,
            7.0,
            7.774597,
            8.0,
            8.225403,
            9.0,
            9.774597,
            10,
        ]

        disc_info = m.t.get_discretization_info()

        self.assertTrue(disc_info['scheme'] == 'LAGRANGE-LEGENDRE')

        for idx, val in enumerate(disc_info['tau_points']):
            self.assertAlmostEqual(val, expected_tau_points[idx])

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])

        m = self.m.clone()
        with self.assertRaises(ValueError):
            disc = TransformationFactory('dae.collocation')
            disc.apply_to(m, ncp=15, scheme='LAGRANGE-LEGENDRE')

        # Restore initial flag value
        pyomo.dae.plugins.colloc.numpy_available = colloc_numpy_avail

    # test discretization using fewer points than ContinuousSet initialized
    # with
    def test_initialized_continuous_set(self):
        m = ConcreteModel()
        m.t = ContinuousSet(initialize=[0, 1, 2, 3, 4])
        m.v = Var(m.t)
        m.dv = DerivativeVar(m.v)

        log_out = StringIO()
        with LoggingIntercept(log_out, 'pyomo.dae'):
            TransformationFactory('dae.collocation').apply_to(m, nfe=2)
        self.assertIn('More finite elements', log_out.getvalue())

    # Test discretizing an invalid derivative
    def test_invalid_derivative(self):
        m = ConcreteModel()
        m.t = ContinuousSet(bounds=(0, 10))
        m.v = Var(m.t)
        m.dv = DerivativeVar(m.v, wrt=(m.t, m.t, m.t))

        with self.assertRaises(DAE_Error):
            TransformationFactory('dae.collocation').apply_to(m)

    # test reduce_collocation_points invalid options
    def test_reduce_colloc_invalid(self):
        m = self.m.clone()
        m.u = Var(m.t)
        m2 = m.clone()

        disc = TransformationFactory('dae.collocation')
        disc2 = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=5, ncp=3)

        # No ContinuousSet specified
        with self.assertRaises(TypeError):
            disc.reduce_collocation_points(m, contset=None)

        # Component passed in is not a ContinuousSet
        with self.assertRaises(TypeError):
            disc.reduce_collocation_points(m, contset=m.s)

        # Call reduce_collocation_points method before applying discretization
        with self.assertRaises(RuntimeError):
            disc2.reduce_collocation_points(m2, contset=m2.t)

        # Call reduce_collocation_points on a ContinuousSet that hasn't been
        #  discretized
        m2.tt = ContinuousSet(bounds=(0, 1))
        disc2.apply_to(m2, wrt=m2.t)
        with self.assertRaises(ValueError):
            disc2.reduce_collocation_points(m2, contset=m2.tt)

        # No Var specified
        with self.assertRaises(TypeError):
            disc.reduce_collocation_points(m, contset=m.t, var=None)

        # Component passed in is not a Var
        with self.assertRaises(TypeError):
            disc.reduce_collocation_points(m, contset=m.t, var=m.s)

        # New ncp not specified
        with self.assertRaises(TypeError):
            disc.reduce_collocation_points(m, contset=m.t, var=m.v1, ncp=None)

        # Negative ncp specified
        with self.assertRaises(ValueError):
            disc.reduce_collocation_points(m, contset=m.t, var=m.v1, ncp=-3)

        # Too large ncp specified
        with self.assertRaises(ValueError):
            disc.reduce_collocation_points(m, contset=m.t, var=m.v1, ncp=10)

        # Passing Vars not indexed by the ContinuousSet
        m.v2 = Var()
        m.v3 = Var(m.s)
        m.v4 = Var(m.s, m.s)

        with self.assertRaises(IndexError):
            disc.reduce_collocation_points(m, contset=m.t, var=m.v2, ncp=1)

        with self.assertRaises(IndexError):
            disc.reduce_collocation_points(m, contset=m.t, var=m.v3, ncp=1)

        with self.assertRaises(IndexError):
            disc.reduce_collocation_points(m, contset=m.t, var=m.v4, ncp=1)

        # Calling reduce_collocation_points more than once
        disc.reduce_collocation_points(m, contset=m.t, var=m.u, ncp=1)
        with self.assertRaises(RuntimeError):
            disc.reduce_collocation_points(m, contset=m.t, var=m.u, ncp=1)

    # test trying to discretize a ContinuousSet twice
    def test_discretize_twice(self):
        m = self.m.clone()

        disc1 = TransformationFactory('dae.collocation')
        disc1.apply_to(m, nfe=5, ncp=3)

        disc2 = TransformationFactory('dae.collocation')

        with self.assertRaises(DAE_Error):
            disc2.apply_to(m, nfe=5, ncp=3)

    # test reduce_collocation_points on var indexed by single ContinuousSet
    def test_reduce_colloc_single_index(self):
        m = self.m.clone()
        m.u = Var(m.t)

        m2 = m.clone()
        m3 = m.clone()

        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=5, ncp=3)
        disc.reduce_collocation_points(m, contset=m.t, var=m.u, ncp=1)

        self.assertTrue(hasattr(m, 'u_interpolation_constraints'))
        self.assertEqual(len(m.u_interpolation_constraints), 10)

        disc2 = TransformationFactory('dae.collocation')
        disc2.apply_to(m2, wrt=m2.t, nfe=5, ncp=3)
        disc2.reduce_collocation_points(m2, contset=m2.t, var=m2.u, ncp=3)

        self.assertFalse(hasattr(m2, 'u_interpolation_constraints'))

        disc3 = TransformationFactory('dae.collocation')
        disc3.apply_to(m3, wrt=m3.t, nfe=5, ncp=3)
        disc3.reduce_collocation_points(m3, contset=m3.t, var=m3.u, ncp=2)

        self.assertTrue(hasattr(m3, 'u_interpolation_constraints'))
        self.assertEqual(len(m3.u_interpolation_constraints), 5)

    # test reduce_collocation_points on var indexed by ContinuousSet and Set
    def test_reduce_colloc_multi_index(self):
        m = self.m.clone()
        m.u = Var(m.t, m.s)

        m2 = m.clone()
        m3 = m.clone()

        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, nfe=5, ncp=3)
        disc.reduce_collocation_points(m, contset=m.t, var=m.u, ncp=1)

        self.assertTrue(hasattr(m, 'u_interpolation_constraints'))
        self.assertEqual(len(m.u_interpolation_constraints), 30)

        disc2 = TransformationFactory('dae.collocation')
        disc2.apply_to(m2, wrt=m2.t, nfe=5, ncp=3)
        disc2.reduce_collocation_points(m2, contset=m2.t, var=m2.u, ncp=3)

        self.assertFalse(hasattr(m2, 'u_interpolation_constraints'))

        disc3 = TransformationFactory('dae.collocation')
        disc3.apply_to(m3, wrt=m3.t, nfe=5, ncp=3)
        disc3.reduce_collocation_points(m3, contset=m3.t, var=m3.u, ncp=2)

        self.assertTrue(hasattr(m3, 'u_interpolation_constraints'))
        self.assertEqual(len(m3.u_interpolation_constraints), 15)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections
import itertools
import operator
import logging

from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    InEnum,
    document_kwargs_from_configdict,
)
from pyomo.common.dependencies import scipy, numpy as np
from pyomo.common.enums import ObjectiveSense
from pyomo.common.gc_manager import PauseGC
from pyomo.common.numeric_types import native_types, value
from pyomo.common.timing import TicTocTimer

from pyomo.core.base import (
    Block,
    Objective,
    Constraint,
    Var,
    Param,
    Expression,
    SortComponents,
    Suffix,
    SymbolMap,
)
from pyomo.opt import WriterFactory
from pyomo.repn.linear import LinearRepnVisitor
from pyomo.repn.linear_template import LinearTemplateRepnVisitor
from pyomo.repn.util import (
    FileDeterminism,
    FileDeterminism_to_SortComponents,
    TemplateVarRecorder,
    categorize_valid_components,
    initialize_var_map_from_column_order,
    ordered_active_constraints,
)

### FIXME: Remove the following as soon as non-active components no
### longer report active==True
from pyomo.core.base import Set, RangeSet, ExternalFunction
from pyomo.network import Port

logger = logging.getLogger(__name__)

RowEntry = collections.namedtuple('RowEntry', ['constraint', 'bound_type'])


# TODO: make a proper base class
class LinearStandardFormInfo:
    """Return type for LinearStandardFormCompiler.write()

    Attributes
    ----------
    c : scipy.sparse.csc_array

        The objective coefficients.  Note that this is a sparse array
        and may contain multiple rows (for multiobjective problems).  The
        objectives may be calculated by ``c @ x``

    c_offset : numpy.ndarray

        The list of objective constant offsets

    A : scipy.sparse.csc_array

        The constraint coefficients.  The constraint bodies may be
        calculated by ``A @ x``

    rhs : numpy.ndarray

        The constraint right-hand sides.

    rows : List[Tuple[ConstraintData, int]]

        The list of Pyomo constraint objects corresponding to the rows
        in `A`.  Each element in the list is a 2-tuple of
        (ConstraintData, row_multiplier).  The `row_multiplier` will be
        +/- 1 indicating if the row was multiplied by -1 (corresponding
        to a constraint lower bound) or +1 (upper bound).

    columns : List[VarData]

        The list of Pyomo variable objects corresponding to columns in
        the `A` and `c` matrices.

    objectives : List[ObjectiveData]

        The list of Pyomo objective objects corresponding to the active objectives

    eliminated_vars: List[Tuple[VarData, NumericExpression]]

        The list of variables from the original model that do not appear
        in the standard form (usually because they were replaced by
        nonnegative variables).  Each entry is a 2-tuple of
        (:py:class:`VarData`, :py:class`NumericExpression`|`float`).
        The list is in the necessary order for correct evaluation (i.e.,
        all variables appearing in the expression must either have
        appeared in the standard form, or appear *earlier* in this list.

    """

    def __init__(self, c, c_offset, A, rhs, rows, columns, objectives, eliminated_vars):
        self.c = c
        self.c_offset = c_offset
        self.A = A
        self.rhs = rhs
        self.rows = rows
        self.columns = columns
        self.objectives = objectives
        self.eliminated_vars = eliminated_vars

    @property
    def x(self):
        "Alias for :attr:`columns`"
        return self.columns

    @property
    def b(self):
        "Alias for :attr:`rhs`"
        return self.rhs


@WriterFactory.register(
    'compile_standard_form',
    r'Compile an LP to standard form (:math:`\min c^Tx s.t. Ax \le b)`',
)
class LinearStandardFormCompiler:
    r"""Compiler to convert an LP to the matrix representation of the
    standard form:

    .. math::

        \min\ & c^Tx \\
        s.t.\ & Ax \le b

    and return the compiled representation as NumPy arrays and SciPy
    sparse matrices.

    """

    CONFIG = ConfigBlock('compile_standard_form')

    CONFIG.declare(
        'nonnegative_vars',
        ConfigValue(
            default=False,
            domain=bool,
            description='Convert all variables to be nonnegative variables',
        ),
    )
    CONFIG.declare(
        'slack_form',
        ConfigValue(
            default=False,
            domain=bool,
            description='Add slack variables and return '
            r':math:`\min c^Tx; s.t. Ax = b`',
        ),
    )
    CONFIG.declare(
        'mixed_form',
        ConfigValue(
            default=False,
            domain=bool,
            description='Return A in mixed form (the comparison operator is a '
            'mix of <=, ==, and >=)',
        ),
    )
    CONFIG.declare(
        'set_sense',
        ConfigValue(
            default=ObjectiveSense.minimize,
            domain=InEnum(ObjectiveSense),
            description='If not None, map all objectives to the specified sense.',
        ),
    )
    CONFIG.declare(
        'show_section_timing',
        ConfigValue(
            default=False,
            domain=bool,
            description='Print timing after each stage of the compilation process',
        ),
    )
    CONFIG.declare(
        'file_determinism',
        ConfigValue(
            default=FileDeterminism.ORDERED,
            domain=InEnum(FileDeterminism),
            description='How much effort to ensure result is deterministic',
            doc="""
            How much effort do we want to put into ensuring the
            resulting matrices are produced deterministically:

               - ``NONE`` (0): None
               - ``ORDERED`` (10): rely on underlying component ordering (default)
               - ``SORT_INDICES`` (20) : sort keys of indexed components
               - ``SORT_SYMBOLS`` (30) : sort keys AND sort names (not
                 declaration order)

            """,
        ),
    )
    CONFIG.declare(
        'row_order',
        ConfigValue(
            default=None,
            description='Preferred constraint ordering',
            doc="""
            List of constraints in the order that they should appear in
            the resulting ``A`` matrix.  Unspecified constraints will
            appear at the end.""",
        ),
    )
    CONFIG.declare(
        'column_order',
        ConfigValue(
            default=None,
            description='Preferred variable ordering',
            doc="""
            List of variables in the order that they should appear in
            the compiled representation.  Unspecified variables will be
            appended to the end of this list.""",
        ),
    )

    def __init__(self):
        self.config = self.CONFIG()

    @document_kwargs_from_configdict(CONFIG)
    def write(self, model, ostream=None, **options):
        """Convert a model to standard form

        Returns
        -------
        LinearStandardFormInfo

        Parameters
        ----------
        model: ConcreteModel
            The concrete Pyomo model to write out.

        ostream:
            This is provided for API compatibility with other writers
            and is ignored here.

        """
        config = self.config(options)

        # Pause the GC, as the walker that generates the compiled LP
        # representation generates (and disposes of) a large number of
        # small objects.
        with PauseGC():
            return _LinearStandardFormCompiler_impl(config).write(model)


class _LinearStandardFormCompiler_impl:
    # Making these methods class attributes so that others can change the hooks
    _get_visitor = LinearRepnVisitor
    _to_vector = None
    _csc_matrix = None
    _csr_matrix = None

    def __init__(self, config):
        self.config = config
        # We defer the first instantiation of these attributes so we do
        # not trigger the numpy / scipy imports when the module is
        # imported
        if _LinearStandardFormCompiler_impl._to_vector is None:
            _LinearStandardFormCompiler_impl._to_vector = np.fromiter
            _LinearStandardFormCompiler_impl._csc_matrix = scipy.sparse.csc_array
            _LinearStandardFormCompiler_impl._csr_matrix = scipy.sparse.csr_array

    def write(self, model):
        timing_logger = logging.getLogger('pyomo.common.timing.writer')
        timer = TicTocTimer(logger=timing_logger)
        with_debug_timing = (
            timing_logger.isEnabledFor(logging.DEBUG) and timing_logger.hasHandlers()
        )

        sorter = FileDeterminism_to_SortComponents(self.config.file_determinism)
        component_map, unknown = categorize_valid_components(
            model,
            active=True,
            sort=sorter,
            valid={
                Block,
                Constraint,
                Var,
                Param,
                Expression,
                # FIXME: Non-active components should not report as Active
                ExternalFunction,
                Set,
                RangeSet,
                Port,
                # TODO: Piecewise, Complementarity
            },
            targets={Suffix, Objective},
        )
        if unknown:
            raise ValueError(
                "The model ('%s') contains the following active components "
                "that the Linear Standard Form compiler does not know how to "
                "process:\n\t%s"
                % (
                    model.name,
                    "\n\t".join(
                        "%s:\n\t\t%s"
                        % (k, "\n\t\t".join(map(operator.attrgetter('name'), v)))
                        for k, v in unknown.items()
                    ),
                )
            )

        self.var_map = var_map = {}
        initialize_var_map_from_column_order(model, self.config, var_map)

        var_recorder = TemplateVarRecorder(var_map, sorter)
        visitor = self._get_visitor({}, var_recorder=var_recorder)
        template_visitor = LinearTemplateRepnVisitor({}, var_recorder=var_recorder)

        timer.toc('Initialized column order', level=logging.DEBUG)

        # We don't export any suffix information to the Standard Form
        #
        if component_map[Suffix]:
            suffixesByName = {}
            for block in component_map[Suffix]:
                for suffix in block.component_objects(
                    Suffix, active=True, descend_into=False, sort=sorter
                ):
                    if not suffix.export_enabled() or not suffix:
                        continue
                    name = suffix.local_name
                    if name in suffixesByName:
                        suffixesByName[name].append(suffix)
                    else:
                        suffixesByName[name] = [suffix]
            for name, suffixes in suffixesByName.items():
                n = len(suffixes)
                plural = 's' if n > 1 else ''
                logger.warning(
                    f"EXPORT Suffix '{name}' found on {n} block{plural}:\n    "
                    + "\n    ".join(s.name for s in suffixes)
                    + "\nStandard Form compiler ignores export suffixes.  Skipping."
                )

        #
        # Process objective
        #
        set_sense = self.config.set_sense
        objectives = []
        for blk in component_map[Objective]:
            objectives.extend(
                blk.component_data_objects(
                    Objective, active=True, descend_into=False, sort=sorter
                )
            )
        obj_nnz = 0
        obj_offset = []
        obj_data = []
        obj_index = []
        obj_index_ptr = [0]
        for obj in objectives:
            if hasattr(obj, 'template_expr'):
                offset, linear_index, linear_data, _, _ = (
                    template_visitor.expand_expression(obj, obj.template_expr())
                )
                N = len(linear_index)
                obj_index.append(linear_index)
                obj_data.append(linear_data)
                obj_offset.append(offset)
            else:
                repn = visitor.walk_expression(obj.expr)
                N = len(repn.linear)
                obj_index.append(map(var_recorder.var_order.__getitem__, repn.linear))
                obj_data.append(repn.linear.values())
                obj_offset.append(repn.constant)

                if repn.nonlinear is not None:
                    raise ValueError(
                        f"Model objective ({obj.name}) contains nonlinear terms that "
                        "cannot be compiled to standard (linear) form."
                    )

            obj_nnz += N
            if set_sense is not None and set_sense != obj.sense:
                obj_data[-1] = -self._to_vector(obj_data[-1], float, N)
                obj_offset[-1] *= -1
            obj_index_ptr.append(obj_index_ptr[-1] + N)
            if with_debug_timing:
                timer.toc('Objective %s', obj, level=logging.DEBUG)

        #
        # Tabulate constraints
        #
        slack_form = self.config.slack_form
        mixed_form = self.config.mixed_form
        if slack_form and mixed_form:
            raise ValueError("cannot specify both slack_form and mixed_form")
        rows = []
        rhs = []
        con_nnz = 0
        con_data = []
        con_index = []
        con_index_ptr = [0]
        last_parent = None
        for con in ordered_active_constraints(model, self.config):
            if with_debug_timing and con._component is not last_parent:
                if last_parent is not None:
                    timer.toc('Constraint %s', last_parent(), level=logging.DEBUG)
                last_parent = con._component

            if hasattr(con, 'template_expr'):
                offset, linear_index, linear_data, lb, ub = (
                    template_visitor.expand_expression(con, con.template_expr())
                )
                N = len(linear_data)
            else:
                # Note: lb and ub could be a number, expression, or None
                lb, body, ub = con.to_bounded_expression()
                if lb.__class__ not in native_types:
                    lb = value(lb)
                if ub.__class__ not in native_types:
                    ub = value(ub)
                repn = visitor.walk_expression(body)
                if repn.nonlinear is not None:
                    raise ValueError(
                        f"Model constraint ({con.name}) contains nonlinear terms that "
                        "cannot be compiled to standard (linear) form."
                    )

                N = len(repn.linear)
                # Pull out the constant: we will move it to the bounds
                offset = repn.constant
                linear_index = map(var_recorder.var_order.__getitem__, repn.linear)
                linear_data = repn.linear.values()

            if lb is None and ub is None:
                # Note: you *cannot* output trivial (unbounded)
                # constraints in matrix format.  I suppose we could add a
                # slack variable, but that seems rather silly.
                continue

            if not N and offset.__class__ in native_types:
                # This is a constant constraint
                # TODO: add a (configurable) feasibility tolerance
                if (lb is None or lb <= offset) and (ub is None or ub >= offset):
                    continue
                raise InfeasibleError(
                    f"model contains a trivially infeasible constraint, '{con.name}'"
                )

            if mixed_form:
                if lb == ub:
                    con_nnz += N
                    rows.append(RowEntry(con, 0))
                    rhs.append(ub - offset)
                    con_data.append(linear_data)
                    con_index.append(linear_index)
                    con_index_ptr.append(con_nnz)
                else:
                    if ub is not None:
                        if lb is not None:
                            linear_index = list(linear_index)
                        con_nnz += N
                        rows.append(RowEntry(con, 1))
                        rhs.append(ub - offset)
                        con_data.append(linear_data)
                        con_index.append(linear_index)
                        con_index_ptr.append(con_nnz)
                    if lb is not None:
                        con_nnz += N
                        rows.append(RowEntry(con, -1))
                        rhs.append(lb - offset)
                        con_data.append(linear_data)
                        con_index.append(linear_index)
                        con_index_ptr.append(con_nnz)
            elif slack_form:
                if lb == ub:  # TODO: add tolerance?
                    rhs.append(ub - offset)
                else:
                    # add slack variable
                    con_nnz += 1
                    v = Var(name=f'_slack_{len(rhs)}', bounds=(None, None))
                    v.construct()
                    if lb is None:
                        rhs.append(ub - offset)
                        v.lb = 0
                    else:
                        rhs.append(lb - offset)
                        v.ub = 0
                        if ub is not None:
                            v.lb = lb - ub
                    var_map[id(v)] = v
                    if var_recorder.var_order is not None:
                        var_recorder.var_order[id(v)] = slack_col = len(
                            var_recorder.var_order
                        )
                    linear_data = list(linear_data)
                    linear_data.append(1)
                    linear_index = list(linear_index)
                    linear_index.append(slack_col)
                con_nnz += N
                rows.append(RowEntry(con, 1))
                con_data.append(linear_data)
                con_index.append(linear_index)
                con_index_ptr.append(con_nnz)
            else:
                if ub is not None:
                    if lb is not None:
                        linear_index = list(linear_index)
                    con_nnz += N
                    rows.append(RowEntry(con, 1))
                    rhs.append(ub - offset)
                    con_data.append(linear_data)
                    con_index.append(linear_index)
                    con_index_ptr.append(con_nnz)
                if lb is not None:
                    con_nnz += N
                    rows.append(RowEntry(con, -1))
                    rhs.append(offset - lb)
                    con_data.append(-np.array(list(linear_data)))
                    con_index.append(linear_index)
                    con_index_ptr.append(con_nnz)

        if with_debug_timing:
            # report the last constraint
            timer.toc('Constraint %s', last_parent(), level=logging.DEBUG)

        # Get the variable list
        columns = list(var_map.values())
        n_cols = len(columns)

        # Convert the compiled data to scipy sparse matrices
        c = self._create_csc(obj_data, obj_index, obj_index_ptr, obj_nnz, n_cols)
        A = self._create_csc(con_data, con_index, con_index_ptr, con_nnz, n_cols)

        if with_debug_timing:
            timer.toc('Formed matrices', level=logging.DEBUG)

        # Some variables in the var_map may not actually appear in the
        # objective or constraints (e.g., added from col_order, or
        # multiplied by 0 in the expressions).  The easiest way to check
        # for empty columns is to convert from CSR to CSC and then look
        # at the index pointer list (an O(num_var) operation).
        c_ip = c.indptr
        A_ip = A.indptr
        active_var_mask = (A_ip[1:] > A_ip[:-1]) | (c_ip[1:] > c_ip[:-1])

        obj_offset = np.array(obj_offset)

        # Note: since we are doing matrix multiplication and then
        # removing the fixed columns, we only want to go through that
        # hassle for columns with actual nonzero entries.
        fixed_vars = (
            np.array([v.fixed for v in columns], dtype=np.bool_) & active_var_mask
        )
        if any(fixed_vars):
            x = np.array([columns[i].value for i, f in enumerate(fixed_vars) if f])
            obj_offset = obj_offset + c[:, fixed_vars] @ x
            rhs = np.array(rhs) - A[:, fixed_vars] @ x
            # Update the active variable mask to exclude the fixed_vars
            # (that we just moved to the RHS)
            active_var_mask &= ~fixed_vars
            c = c[:, active_var_mask]
            A = A[:, active_var_mask]
            columns = [v for k, v in zip(active_var_mask, columns) if k]
            n_cols = len(columns)
        else:
            # Masks on NumPy arrays are very fast.  Build the reduced A
            # indptr and then check if we actually have to manipulate the
            # columns
            augmented_mask = np.concatenate((active_var_mask, [True]))
            reduced_A_indptr = A.indptr[augmented_mask]
            n_cols -= len(reduced_A_indptr) - 1
            if n_cols > 0:
                columns = [v for k, v in zip(active_var_mask, columns) if k]
                c = self._csc_matrix(
                    (c.data, c.indices, c.indptr[augmented_mask]),
                    [c.shape[0], len(columns)],
                )
                # active_var_idx[-1] = len(columns)
                A = self._csc_matrix(
                    (A.data, A.indices, reduced_A_indptr), [A.shape[0], len(columns)]
                )

        if with_debug_timing:
            timer.toc('Eliminated %s unused columns', n_cols, level=logging.DEBUG)

        if self.config.nonnegative_vars:
            c, A, columns, eliminated_vars = self._csc_to_nonnegative_vars(
                c, A, columns
            )
        else:
            eliminated_vars = []

        info = LinearStandardFormInfo(
            c, obj_offset, A, rhs, rows, columns, objectives, eliminated_vars
        )
        timer.toc("Generated linear standard form representation", delta=False)
        return info

    def _create_csc(self, data, index, index_ptr, nnz, n_cols):
        data = self._to_vector(itertools.chain.from_iterable(data), np.float64, nnz)
        index = self._to_vector(itertools.chain.from_iterable(index), np.int32, nnz)
        index_ptr = np.array(index_ptr, dtype=np.int32)

        if not nnz:
            # The empty CSC has no (or few) rows and a large number of
            # columns and no nonzeros: it is faster / easier to create
            # the empty CSR on the python side and convert it to CSC on
            # the C (numpy) side, as opposed to creating the large [0] *
            # (n_cols + 1) array on the Python side and transfer it to C
            # (numpy)
            return self._csr_matrix(
                (data, index, index_ptr), [len(index_ptr) - 1, n_cols]
            ).tocsc()

        A = self._csr_matrix((data, index, index_ptr), [len(index_ptr) - 1, n_cols])
        A = A.tocsc()
        A.sum_duplicates()
        A.eliminate_zeros()
        return A

    def _csc_to_nonnegative_vars(self, c, A, columns):
        eliminated_vars = []
        new_columns = []
        new_c_data = []
        new_c_indices = []
        new_c_indptr = [0]
        new_A_data = []
        new_A_indices = []
        new_A_indptr = [0]
        for i, v in enumerate(columns):
            lb, ub = v.bounds
            if lb is None or lb < 0:
                name = v.name
                new_columns.append(
                    Var(
                        name=f'_neg_{i}',
                        domain=v.domain,
                        bounds=(0, None if lb is None else -lb),
                    )
                )
                new_columns[-1].construct()
                s, e = A.indptr[i : i + 2]
                new_A_data.append(-A.data[s:e])
                new_A_indices.append(A.indices[s:e])
                new_A_indptr.append(new_A_indptr[-1] + e - s)
                s, e = c.indptr[i : i + 2]
                new_c_data.append(-c.data[s:e])
                new_c_indices.append(c.indices[s:e])
                new_c_indptr.append(new_c_indptr[-1] + e - s)
                if ub is None or ub > 0:
                    # Crosses 0; split into 2 vars
                    new_columns.append(
                        Var(name=f'_pos_{i}', domain=v.domain, bounds=(0, ub))
                    )
                    new_columns[-1].construct()
                    s, e = A.indptr[i : i + 2]
                    new_A_data.append(A.data[s:e])
                    new_A_indices.append(A.indices[s:e])
                    new_A_indptr.append(new_A_indptr[-1] + e - s)
                    s, e = c.indptr[i : i + 2]
                    new_c_data.append(c.data[s:e])
                    new_c_indices.append(c.indices[s:e])
                    new_c_indptr.append(new_c_indptr[-1] + e - s)
                    eliminated_vars.append((v, new_columns[-1] - new_columns[-2]))
                else:
                    new_columns[-1].lb = -ub
                    eliminated_vars.append((v, -new_columns[-1]))
            else:  # lb >= 0
                new_columns.append(v)
                s, e = A.indptr[i : i + 2]
                new_A_data.append(A.data[s:e])
                new_A_indices.append(A.indices[s:e])
                new_A_indptr.append(new_A_indptr[-1] + e - s)
                s, e = c.indptr[i : i + 2]
                new_c_data.append(c.data[s:e])
                new_c_indices.append(c.indices[s:e])
                new_c_indptr.append(new_c_indptr[-1] + e - s)

        n_cols = len(new_columns)
        c = self._csc_matrix(
            (np.concatenate(new_c_data), np.concatenate(new_c_indices), new_c_indptr),
            [c.shape[0], n_cols],
        )
        A = self._csc_matrix(
            (np.concatenate(new_A_data), np.concatenate(new_A_indices), new_A_indptr),
            [A.shape[0], n_cols],
        )
        return c, A, new_columns, eliminated_vars



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import platform
import pyomo.environ as pyo
import pyomo.common.unittest as unittest
from pyomo.common.fileutils import find_library
from pyomo.opt import check_available_solvers
from pyomo.core.base.external import nan

flib = find_library("asl_external_demo")
is_pypy = platform.python_implementation().lower().startswith('pypy')


@unittest.skipUnless(flib, 'Could not find the "asl_external_demo.so" library')
class TestAMPLExternalFunction(unittest.TestCase):
    @unittest.skipIf(is_pypy, 'Cannot evaluate external functions under pypy')
    def test_eval_function(self):
        m = pyo.ConcreteModel()
        m.tf = pyo.ExternalFunction(library=flib, function="demo_function")
        self.assertAlmostEqual(m.tf("sum", 1, 2, 3)(), 6, 4)
        self.assertAlmostEqual(m.tf("inv", 1, 2, 3)(), 1.8333333, 4)
        m.cbrt = pyo.ExternalFunction(library=flib, function="safe_cbrt")
        self.assertAlmostEqual(m.cbrt(6)(), 1.81712059, 4)
        self.assertStructuredAlmostEqual(
            m.cbrt.evaluate_fgh([0]), (0, [100951], [-1.121679e13]), reltol=1e-5
        )

    @unittest.skipIf(is_pypy, 'Cannot evaluate external functions under pypy')
    def test_eval_function_fgh(self):
        m = pyo.ConcreteModel()
        m.tf = pyo.ExternalFunction(library=flib, function="demo_function")

        f, g, h = m.tf.evaluate_fgh(("sum", 1, 2, 3))
        self.assertEqual(f, 6)
        self.assertEqual(g, [nan, 1, 1, 1])
        self.assertEqual(h, [nan, nan, 0, nan, 0, 0, nan, 0, 0, 0])

        f, g, h = m.tf.evaluate_fgh(("inv", 1, 2, 3))
        self.assertAlmostEqual(f, 1.8333333, 4)
        self.assertStructuredAlmostEqual(g, [nan, -1, -1 / 4, -1 / 9])
        self.assertStructuredAlmostEqual(
            h, [nan, nan, 2, nan, 0, 1 / 4, nan, 0, 0, 2 / 27]
        )

    @unittest.skipUnless(
        check_available_solvers('ipopt'), "The 'ipopt' solver is not available"
    )
    def test_solve_function(self):
        m = pyo.ConcreteModel()
        m.sum_it = pyo.ExternalFunction(library=flib, function="demo_function")
        m.cbrt = pyo.ExternalFunction(library=flib, function="safe_cbrt")
        m.x = pyo.Var(initialize=4.0)
        m.y = pyo.Var(initialize=0)
        m.y.fix()
        # Note: this also tests passing constant expressions to external
        # functions in the NL writer
        m.c = pyo.Constraint(expr=1.5 == m.sum_it("inv", 3, m.x, 1 / (m.y + 1)))
        m.o = pyo.Objective(expr=m.cbrt(m.x))
        solver = pyo.SolverFactory("ipopt")
        solver.solve(m, tee=True)
        self.assertAlmostEqual(m.x(), 6, 4)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""This module contains functions to interrogate the size of a Pyomo model."""

import logging

from pyomo.common.collections import ComponentSet, Bunch
from pyomo.core import Block, Constraint, Var
import pyomo.core.expr as EXPR
from pyomo.gdp import Disjunct, Disjunction

default_logger = logging.getLogger('pyomo.util.model_size')
default_logger.setLevel(logging.INFO)


class ModelSizeReport(Bunch):
    """Stores model size information.

    Activated blocks are those who have an active flag of True and whose
    parent, if exists, is an activated block or an activated Disjunct.

    Activated constraints are those with an active flag of True and: are
    reachable via an activated Block, are on an activated Disjunct, or are on a
    disjunct with indicator_var fixed to 1 with active flag True.

    Activated variables refer to the presence of the variable on an activated
    constraint, or that the variable is an indicator_var for an activated
    Disjunct.

    Activated disjuncts refer to disjuncts with an active flag of True, have an
    unfixed indicator_var, and who participate in an activated Disjunction.

    Activated disjunctions follow the same rules as activated constraints.

    """

    pass


def build_model_size_report(model):
    """Build a model size report object."""
    report = ModelSizeReport()
    activated_disjunctions = ComponentSet()
    activated_disjuncts = ComponentSet()
    fixed_true_disjuncts = ComponentSet()
    activated_constraints = ComponentSet()
    activated_vars = ComponentSet()
    new_containers = (model,)

    while new_containers:
        new_activated_disjunctions = ComponentSet()
        new_activated_disjuncts = ComponentSet()
        new_fixed_true_disjuncts = ComponentSet()
        new_activated_constraints = ComponentSet()

        for container in new_containers:
            (
                next_activated_disjunctions,
                next_fixed_true_disjuncts,
                next_activated_disjuncts,
                next_activated_constraints,
            ) = _process_activated_container(container)
            new_activated_disjunctions.update(next_activated_disjunctions)
            new_activated_disjuncts.update(next_activated_disjuncts)
            new_fixed_true_disjuncts.update(next_fixed_true_disjuncts)
            new_activated_constraints.update(next_activated_constraints)

        new_containers = (new_activated_disjuncts - activated_disjuncts) | (
            new_fixed_true_disjuncts - fixed_true_disjuncts
        )

        activated_disjunctions.update(new_activated_disjunctions)
        activated_disjuncts.update(new_activated_disjuncts)
        fixed_true_disjuncts.update(new_fixed_true_disjuncts)
        activated_constraints.update(new_activated_constraints)

    activated_vars.update(
        var
        for constr in activated_constraints
        for var in EXPR.identify_variables(constr.body, include_fixed=False)
    )
    activated_vars.update(
        disj.indicator_var.get_associated_binary() for disj in activated_disjuncts
    )

    report.activated = Bunch()
    report.activated.variables = len(activated_vars)
    report.activated.binary_variables = sum(1 for v in activated_vars if v.is_binary())
    report.activated.integer_variables = sum(
        1 for v in activated_vars if v.is_integer() and not v.is_binary()
    )
    report.activated.continuous_variables = sum(
        1 for v in activated_vars if v.is_continuous()
    )
    report.activated.disjunctions = len(activated_disjunctions)
    report.activated.disjuncts = len(activated_disjuncts)
    report.activated.constraints = len(activated_constraints)
    report.activated.nonlinear_constraints = sum(
        1 for c in activated_constraints if c.body.polynomial_degree() not in (1, 0)
    )

    report.overall = Bunch()
    block_like = (Block, Disjunct)
    all_vars = ComponentSet(model.component_data_objects(Var, descend_into=block_like))
    report.overall.variables = len(all_vars)
    report.overall.binary_variables = sum(1 for v in all_vars if v.is_binary())
    report.overall.integer_variables = sum(
        1 for v in all_vars if v.is_integer() and not v.is_binary()
    )
    report.overall.continuous_variables = sum(1 for v in all_vars if v.is_continuous())
    report.overall.disjunctions = sum(
        1 for d in model.component_data_objects(Disjunction, descend_into=block_like)
    )
    report.overall.disjuncts = sum(
        1 for d in model.component_data_objects(Disjunct, descend_into=block_like)
    )
    report.overall.constraints = sum(
        1 for c in model.component_data_objects(Constraint, descend_into=block_like)
    )
    report.overall.nonlinear_constraints = sum(
        1
        for c in model.component_data_objects(Constraint, descend_into=block_like)
        if c.body.polynomial_degree() not in (1, 0)
    )

    report.warning = Bunch()
    report.warning.unassociated_disjuncts = sum(
        1
        for d in model.component_data_objects(Disjunct, descend_into=block_like)
        if not d.indicator_var.fixed and d not in activated_disjuncts
    )

    return report


def log_model_size_report(model, logger=default_logger):
    """Generate a report logging the model size."""
    logger.info(build_model_size_report(model))


def _process_activated_container(blk):
    """Process a container object, returning the new components found."""
    new_fixed_true_disjuncts = ComponentSet(
        disj
        for disj in blk.component_data_objects(Disjunct, active=True)
        if disj.indicator_var.value and disj.indicator_var.fixed
    )
    new_activated_disjunctions = ComponentSet(
        blk.component_data_objects(Disjunction, active=True)
    )
    new_activated_disjuncts = ComponentSet(
        disj
        for disjtn in new_activated_disjunctions
        for disj in _activated_disjuncts_in_disjunction(disjtn)
    )
    new_activated_constraints = ComponentSet(
        blk.component_data_objects(Constraint, active=True)
    )
    return (
        new_activated_disjunctions,
        new_fixed_true_disjuncts,
        new_activated_disjuncts,
        new_activated_constraints,
    )


def _activated_disjuncts_in_disjunction(disjtn):
    """Retrieve generator of activated disjuncts on disjunction."""
    return (
        disj
        for disj in disjtn.disjuncts
        if disj.active and not disj.indicator_var.fixed
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.tempfiles import TempfileManager
from pyomo.opt import TerminationCondition


def ipopt_solve_with_stats(model, solver, max_iter=500, max_cpu_time=120):
    """
    Run the solver (must be ipopt) and return the convergence statistics

    Parameters
    ----------
    model : Pyomo model
       The pyomo model to be solved

    solver : Pyomo solver
       The pyomo solver to use - it must be ipopt, but with whichever options are preferred

    max_iter : int
       The maximum number of iterations to allow for ipopt

    max_cpu_time : int
       The maximum cpu time to allow for ipopt (in seconds)

    Returns
    -------
       Returns a tuple with (solve status object, bool (solve successful or not), number of iters, solve time, regularization value at solution)
    """
    # ToDo: Check that the "solver" is, in fact, IPOPT

    TempfileManager.push()
    tempfile = TempfileManager.create_tempfile(suffix='ipopt_out', text=True)
    opts = {'output_file': tempfile, 'max_iter': max_iter, 'max_cpu_time': max_cpu_time}

    status_obj = solver.solve(model, options=opts, tee=True)
    solved = True
    if status_obj.solver.termination_condition != TerminationCondition.optimal:
        solved = False

    iters = 0
    time = 0
    line_m_2 = None
    line_m_1 = None
    # parse the output file to get the iteration count, solver times, etc.
    with open(tempfile, 'r') as f:
        for line in f:
            if line.startswith('Number of Iterations....:'):
                tokens = line.split()
                iters = int(tokens[3])
                tokens_m_2 = line_m_2.split()
                regu = str(tokens_m_2[6])
            elif line.startswith(
                'Total CPU secs in IPOPT (w/o function evaluations)   ='
            ):
                tokens = line.split()
                time += float(tokens[9])
            elif line.startswith(
                'Total CPU secs in NLP function evaluations           ='
            ):
                tokens = line.split()
                time += float(tokens[8])
            line_m_2 = line_m_1
            line_m_1 = line

    TempfileManager.pop(remove=True)
    return status_obj, solved, iters, time, regu
